import unittest
from hein_control.timepoint import TimePoint, ActionTimePoint
from .imported_methods import fifty_ms, one_second

tp_name = f'test 1'
tp_desc = 'test case time point'


class TestTimePoint(unittest.TestCase):
    def setUp(self) -> None:
        self.tp_one = TimePoint(
            1.,
            name=tp_name,
            description=tp_desc
        )

    def test_class(self):
        """tests basic functionality of the class"""
        self.assertEqual(
            self.tp_one.name,
            tp_name
        )
        self.assertEqual(
            1.,
            self.tp_one.time_delta,
        )
        self.assertEqual(
            tp_desc,
            self.tp_one.description
        )

    def test_instance_retrieval(self):
        """tests instance retrieval method"""
        self.assertTrue(
            self.tp_one is TimePoint.get_instance(self.tp_one._uuid)
        )
        self.assertTrue(
            self.tp_one is TimePoint.get_instance(self.tp_one.uuid)
        )

    def test_magic_methods(self):
        """tests comparison methods"""
        self.assertTrue(
            self.tp_one < TimePoint(2)
        )
        self.assertTrue(
            self.tp_one <= TimePoint(2)
        )
        self.assertTrue(
            self.tp_one > TimePoint(0.5)
        )
        self.assertTrue(
            self.tp_one >= TimePoint(0.5)
        )
        self.assertTrue(
            self.tp_one == TimePoint(1)
        )


class TestActionTimePoint(unittest.TestCase):
    def test_sequencing(self):
        atp = ActionTimePoint(
            3.,
        )
        atp.add_action(fifty_ms)
        atp.trigger()
        atp.wait_for_completion()
        self.assertRaises(
            SystemError,
            atp.trigger,
        )

        # todo better test for these retrievals
        durations = atp.action_durations
        starts = atp.action_start_times
        ends = atp.action_completion_times
        action_returns = atp.action_returns
        timestamps = atp.started_timestamps
        dct = atp.as_dict()

    def test_flags(self):
        atp = ActionTimePoint(3)
        atp.add_action(one_second)
        atp.trigger()
        self.assertTrue(
            atp.in_progress
        )
        self.assertFalse(
            atp.complete
        )
        atp.wait_for_completion()
        self.assertFalse(atp.in_progress)
        self.assertTrue(atp.complete)



import logging
from typing import List, Callable, Union
from .action import ConfiguredAction, TrackedAction


logger = logging.getLogger(__name__)


class Step(ConfiguredAction):
    def __init__(self,
                 action: Union[Callable, ConfiguredAction],
                 next_step: 'Step' = None,
                 bypass_return: bool = None,
                 **kwargs,
                 ):
        """
        A step in an automation sequence. An instance of a Step should not be created, instead one of the subclasses
        should be instantiated.

        A Step can be bypassed (skipped over).

        A `Step` will have a `next_step`, which is another `Step` or `None`; in this way,
        steps in an `Automation` can be linked, and different work flows can be created.

        A `Step` is also callable. The details of this will depend on the specific subclass.

        :param callable, action: a callable, most of the time it will be a function, that will be executed when the
            Step is called/run
        :param Step, next_step: the Step that should follow after this Step has been run
        :param bypass_return: for a CustomStep, there might be some return value. If a Step is bypassed but you need
            a value to be returned still, you can hardcode a value to be returned. For example, if the step is to
            take a measurement with an  instrument and update a value, you can set what the measurement value that
            gets returned should be here
        :param dict, kwargs: dictionary must contain entries with keys that are identical to all the parameters for
            the callable action
        """
        super().__init__(
            action=action,
            **kwargs,
        )
        # storage list for executions
        self._executions: List[TrackedAction] = []
        # todo consider using a class-level logger rather than childing for each instance
        self.logger = logger.getChild(f'{self.__class__.__name__} {self.name}')

        self._next_step: Step = None  # the step that follows from this step
        self.next_step = next_step
        self._bypass_step: bool = False  # Bool, if set to True, don't run what the step should do aka bypass it

        self._bypass_return = None
        self.bypass_return = bypass_return

    def __call__(self, **kwargs) -> Union[dict, bool, None]:
        """

        If the Step will not be bypassed, create a TrackedAction instance for the action and execute it, and return
        the correct return for the specific Step that is being executed; CustomStep can return dict or None,
        ConditionCheck returns bool, and IfStep returns None.

        :param dict, kwargs: dictionary must contain entries with keys that are identical to all the parameters for
            the callable action, and the values are the actual values/items for the corresponding keys
        :return:
        """
        if self._bypass_step is True:
            self.logger.debug(f'bypass step')
            action_return = self._bypass_return
        else:
            self._executions.append(
                TrackedAction.create_from_configured(self.action)
            )
            self._executions[-1].__call__(**kwargs)
            action_return = self.last_action_return()

        return action_return

    @property
    def executions(self) -> List[TrackedAction]:
        """previous executions of this Step"""
        return self._executions

    def last_action_return(self) -> Union[dict, bool, None]:
        return self._executions[-1]._action_return

    @property
    def next_step(self) -> 'Step':
        """The step that follows this step"""
        return self._next_step

    @next_step.setter
    def next_step(self,
                  value: 'Step'
                  ):
        if value is not None:
            if isinstance(value, Step) is False:
                raise TypeError('value must be of type Step')
            self._next_step = value

    @next_step.deleter
    def next_step(self,
                  ):
        self._next_step = None

    @property
    def bypass_step(self) -> bool:
        """whether the step should be bypassed"""
        return self._bypass_step

    @bypass_step.setter
    def bypass_step(self,
                    value: bool
                    ):
        if type(value)is not bool:
            raise TypeError('value must be of type bool')
        self._bypass_step = value
        self.logger.debug(f'set bypass step to True')

    @bypass_step.deleter
    def bypass_step(self):
        self._bypass_step = False

    @property
    def bypass_return(self):
        """
        If a Step will be bypassed but normally the action that should be executed needs to return some
        value, then some default/hardcoded value can be set to be returned. The type of value will depending on
        the specific subclass of Step that is being bypassed
        """
        return self._bypass_return

    @bypass_return.setter
    def bypass_return(self,
                      value):
        self._bypass_return = value

    @bypass_return.deleter
    def bypass_return(self):
        self._bypass_return = None


class CustomStep(Step):
    def __init__(self,
                 action: Callable,
                 next_step: Step = None,
                 **kwargs,
                 ):
        """
         A `Step` that could have a return value after the callable has been executed; the return type must be a
        dictionary where keys are argument names (identical to the arguments of the parameters of the callable and
        therefore also identical to the keys in the dictionary used when an `Automation` is instantiated),
        and the values are the new value they should take. See the `Automation` class as to why this is important.

        :param callable, action: a callable, most of the time it will be a function, that will be executed when the
            Step is called/run. It may return a dictionary.
        :param Step, next_step: the Step that should follow after this Step has been run
        :param bypass_return: there might be some return value. If a CustomStep is bypassed but you need
            a value to be returned still, you can hardcode a value to be returned. For example, if the step is to
            take a measurement with an  instrument and update a value, you can set what the measurement value that
            gets returned should be here
        :param dict, kwargs: dictionary must contain entries with keys that are identical to all the parameters for
            the callable action
        """
        super().__init__(
            action=action,
            next_step=next_step,
            **kwargs,
        )

    def __call__(self, **kwargs) -> Union[dict, None]:
        """

        :param dict, kwargs: dictionary must contain entries with keys that are identical to all the parameters for
            the callable action, and the values are the actual values/items for the corresponding keys
        :return: dict, a dictionary of where entries are keys that are identical to all the parameters for
            the callable action, and the values are the updated value of the corresponding keys
        """
        action_return = super().__call__(**kwargs)
        return action_return


class ConditionCheck(Step):
    def __init__(self,
                 condition: Callable,
                 next_step: Step,
                 **kwargs,
                 ):
        """
        A `Step` that, when executed, returns a `bool`. It is used by the `IfStep` to know what the `next_step` in an
        `Automation` instance should be.

        :param callable, condition: a callable, most of the time it will be a function, that will be executed when the
            Check is called/run. it must return a bool
        :param Step, next_step: the Step that should follow after this Step has been run
        :param kwargs: dictionary must contain entries with keys that are identical to all the parameters for
            the callable action
        """

        super().__init__(
            action=condition,
            next_step=next_step,
            **kwargs,
        )

    def __call__(self, **kwargs) -> bool:
        """
        Run the callable action and return the return of that action; the action should return a bool

        :param dict, kwargs: dictionary must contain entries with keys that are identical to all the parameters for
            the callable action, and the values are the actual values/items for the corresponding keys
        :return: bool, value is based on the return of the callable action
        """
        action_return = super().__call__(**kwargs)
        return action_return


class IfStep(Step):
    def __init__(self,
                 condition_check_list: list = None,
                 **kwargs,
                 ):
        """
         A `Step` in an automation sequence that has multiple possible next steps, which is decided based on a list of
        `ConditionChecks`. If the return of the first `ConditionCheck` is `True`, then the next step of the `IfStep` is
        the `next_step` of the `ConditionCheck`. If the bool is `False`, then iterate through the list of
        `ConditionChecks` until a condition returns `True`. If none of them return `True`, then the last
        `ConditionCheck` in the list is assumed to be `True` (essentially it is like the `else` in an
        `if-elif-...-else` block).

        :param list, condition_check_list: a list of ConditionCheck instances
        :param kwargs: dictionary must contain entries with keys that are identical to all the parameters for
            the callable action
        """
        self._condition_check_list = condition_check_list

        super().__init__(
            action=self.run_conditions,
            next_step=None,
            **kwargs,
        )

    def __call__(self, **kwargs) -> None:
        """

        :param dict, kwargs: dictionary must contain entries with keys that are identical to all the parameters for
            the callable action, and the values are the actual values/items for the corresponding keys
        :return:
        """
        super().__call__(**kwargs)
        return None

    @property
    def condition_check_list(self) -> List[Step]:
        """
        a list of ConditionChecks to iterate through to know what to put as the self.next_step
        when the IfStep instance is run
        """
        return self._condition_check_list

    @condition_check_list.setter
    def condition_check_list(self,
                             value: List[Step],
                             ):
        if isinstance(value, list) is False:
            raise TypeError('value should be of type list')
        elif any([isinstance(step, Step) is False for step in value]):
            raise TypeError('A list of Step instances must be provided')
        self._condition_check_list = value

    @condition_check_list.deleter
    def condition_check_list(self):
        self._condition_check_list = None

    def run_conditions(self,
                       **kwargs,
                       ) -> None:
        """
        Go through each ConditionCheck in the self._condition_check_list to find out what the next_step from this
        IfStep instance should be. If the Step has been set to be bypassed, then the next_step is based on the last
        ConditionCheck in the list

        :param dict, kwargs:  dictionary must contain entries with keys that are identical to all the parameters for
            all the ConditionChecks
        :return:
        """

        if self._bypass_step is False:
            for condition_check in self._condition_check_list:
                if condition_check(**kwargs):
                    resulting_step = condition_check.next_step
                    self.logger.debug(f'true condition is {condition_check}')
                    self._next_step = resulting_step
                    return

        # if None of the condition checks return true or bypass is true, then make the next step by default the
        # resulting step of the last condition step in the list
        last_condition_step = self._condition_check_list[-1]
        resulting_step = last_condition_step.next_step
        self._next_step = resulting_step
        self.logger.debug(f'no condition was true.')




import datetime
import uuid
import threading
import logging
import time
from typing import Union, List, Dict, Callable
from .action import TrackedAction, Action, ConfiguredAction


logger = logging.getLogger(__name__)


class TimePoint(object):
    UUID_PREFIX = 'tp-'

    # timepoint instance tracking
    _instances: List["TimePoint"] = []

    def __init__(self,
                 time_delta: float,
                 name: str = None,
                 description: str = None,
                 parent_sequence_id: str = None,
                 ):
        """
        A class for describing and tracking an event time point

        :param time_delta: time delta for the time point (relative to some other time)
        :param name: convenience name for time point differentiation
        :param description: extended description for differentiation
        :param parent_sequence_id: uuid of parent SamplingScheduler (if applicable)
        """
        # time delta from trigger time
        self.time_delta: float = float(time_delta)
        # storage for start time
        self.sequence_time_started: float = None
        # name for convenience
        self.name: str = name
        # description
        self.description: str = description
        # parent uuid
        self.parent_sequence_id: str = parent_sequence_id

        self._uuid = str(uuid.uuid4())  # assign unique identifier
        TimePoint._instances.append(self)
        # todo add description attribute
        # todo create association between time point and parent scheduler
        # todo finish refactor
        #   - move as_dict, update as necessary

    def __str__(self):
        return (
            f'{self.__class__.__name__} {self.time_delta} s '
        )

    def __repr__(self):
        return self.__str__()

    @property
    def uuid(self) -> str:
        """uuid for the time point"""
        return self.UUID_PREFIX + self._uuid

    @property
    def basic_dict(self) -> dict:
        """basic dictionary for any TimePoint class or subclass"""
        return {
            'time_delta': self.time_delta,
            'uuid': self.uuid,
            'name': self.name,
            'class': self.__class__.__name__,
        }

    @classmethod
    def get_instance(cls, key: str) -> "TimePoint":
        """
        Retreives a timepoint instance by key. The key may be a uuid, or class-uuid

        :param key: key to search by
        :return: TimePoint (or subclass) instance
        """
        for instance in cls._instances:
            if any([
                # instance.name == key,  # todo consider reimplementing
                instance._uuid == key,
                instance.uuid == key,
            ]):
                return instance
        raise KeyError(f'the key "{key}" does not match an identifier for any {cls.__class__.__name__} instances')

    def as_dict(self) -> dict:
        """Returns a key: value representation of the instance"""
        return self.basic_dict

    def sample_time_from_reference(self, reference_time: float):
        """
        Returns the sample time for the time point from the reference time (in seconds)

        :param reference_time: reference time to calculate from
        :return: time to trigger (seconds)
        """
        return self.time_delta + reference_time

    def __lt__(self, other: Union[float, "TimePoint"]):
        if isinstance(other, TimePoint):
            other = other.time_delta
        return self.time_delta < other

    def __eq__(self, other: Union[float, "TimePoint"]):
        if isinstance(other, TimePoint):
            other = other.time_delta
        return self.time_delta == other

    def __gt__(self, other: Union[float, "TimePoint"]):
        if isinstance(other, TimePoint):
            other = other.time_delta
        return self.time_delta > other

    def __le__(self, other: Union[float, "TimePoint"]):
        return self.__lt__(other) or self.__eq__(other)

    def __ge__(self, other: Union[float, "TimePoint"]):
        return self.__gt__(other) or self.__eq__(other)


class ActionTimePoint(TimePoint):
    UUID_PREFIX = 'atp-'

    def __init__(self,
                 time_delta: float,
                 name: str = None,
                 actions: List[Union[Callable, Action, ConfiguredAction]] = [],
                 **action_kwargs,
                 ):
        """
        A class for describing and tracking an event time point where actions (e.g. sampling) are to be triggered

        :param time_delta: time delta for the time point (relative to some other time)
        :param name: convenience name for time point differentiation
        :param actions: action methods to call. These methods will be called in the order provided
        :param action_kwargs: action kwargs to associate with list of actions (primary keys must map to function names)
        """
        # todo create abstract class with context
        #   - individual methods for pre-step, main, and post-step (e.g. turn on pump, run experiment, process data)
        # todo associate timepoint with parent scheduler instance (by UUID)
        # todo create passthrough for sequence_time_started on init
        TimePoint.__init__(
            self,
            time_delta=time_delta,
            name=name,
        )
        # store actions for timepoint
        self._actions: List[TrackedAction] = [
            TrackedAction(action) for action in actions
        ]
        # update kwargs
        for action_name in action_kwargs:
            for action in self._actions:
                if action.name == action_name:
                    action.update_kwargs(**action_kwargs[action_name])
        # internal storage attributes
        self._triggered = False
        self._thread: threading.Thread = None
        self._lock: threading.Lock = threading.Lock()
        # todo create attribute for retrieving action returns

    def __str__(self):
        # todo add attribute for retrieving formatted time (HH:MM:SS)
        out = f'{self.__class__.__name__} {self.string_time_delta}'
        if self.complete is True:
            out += ' COMPLETE'
        elif self.triggered is True:
            out += ' TRIGGERED'
        return out

    def __call__(self, **kwargs):
        self.trigger(**kwargs)

    @property
    def actions(self) -> List[TrackedAction]:
        """actions to be run when the timepoint is called"""
        return self._actions

    def add_action(self,
                   action: Union[str, Callable, Action, ConfiguredAction],
                   location: int = None,
                   ):
        """
        Adds an action to the timepoint. Actions will be executed in the order they appear in the list of actions for
        the ActionTimePoint instance.

        :param action: action to perform. Actions may be added by providing a callable method or name of a predefined
            callable method
        :param location: index location to insert. If left as None the action will be appended
        """
        if location is None:
            location = len(self._actions)
        self._actions.insert(
            location,
            TrackedAction(action),
        )

    def remove_action(self,
                      action: Union[str, Callable, Action],
                      ):
        """
        Removes an action from the list of actions associated with a timepoint.

        :param action: action identifier
        :return:
        """
        raise NotImplementedError()
        # todo implement
        #   - determine identifier
        #   - remove from _actions list
        # todo implement ability to change action associated with time points
        #   - e.g. switch from no pump ON/OFF to method with pump on/off

    def _execute_actions(self, **kwargs):
        """
        Executes the actions defined in the instance.

        :param kwargs: keyword arguments for those action methods
        """
        execution_logger = logger.getChild(f'{self.__class__.__name__} {self.string_time_delta}')
        with self._lock:
            execution_logger.debug('initiating action sequence')
            # for each action, retrieve keyword arguments if specified then execute
            for action in self.actions:
                action_kwargs = {}
                if action.name in kwargs:
                    action_kwargs.update(kwargs[action.name])
                execution_logger.debug(f'executing {action.name}')
                action(**action_kwargs)
                execution_logger.debug(f'{action.name} complete')
            execution_logger.debug('action sequence complete')

    def trigger(self,
                wait: bool = None,
                **kwargs,
                ):
        """
        Triggers the actions of the ActionTimePoint. Keyword arguments will be passed through to their respective methods.

        e.g. if there is a method named "do_this_thing" which expects the keyword argument "value", the expected syntax
        of kwargs would be

            {
                "do_this_thing": {"value": 42},
            }

        :param wait: whether to wait for completion before returning
        :param kwargs: keyword arguments for the actions. Keyword arguments will be passed through to their respective
            methods.
        """
        if self.triggered is True:
            raise SystemError(f'Multiple executions of the same {self.__class__.__name__} are not supported')
        # todo catch multiple calls and raise
        self._triggered = True  # set flag as triggered
        # set up thread
        self._thread = threading.Thread(
            target=self._execute_actions,
            name=f'{str(self)} action thread',
            daemon=True,
            kwargs=kwargs,
        )
        self._thread.start()  # start thread
        if wait is True:  # wait if specified
            self.wait_for_completion()

    @property
    def string_time_delta(self) -> str:
        """formatted string version of time delta"""
        return str(
            datetime.timedelta(seconds=self.time_delta)
        )

    @property
    def triggered(self) -> bool:
        """whether the time point has been triggered"""
        return self._triggered

    @property
    def in_progress(self) -> bool:
        """
        whether the actions are in progress
        (time point is triggered and actions are in the process of being executed)
        """
        if self.triggered is True:
            return self._lock.locked()
        return False

    @property
    def complete(self) -> bool:
        """whether the actions have completed"""
        if self.triggered is True:  # if instance has been triggered, check lock state
            return not self._lock.locked()
        return False

    @property
    def action_start_times(self) -> dict:
        """times when the prescribed actions were started"""
        if self.triggered is True:
            return {
                action.name: action.time_started for action in self.actions
            }

    @property
    def action_completion_times(self) -> dict:
        """times when the prescribed actions were completed"""
        if self.triggered is True:
            return {
                action.name: action.time_completed for action in self.actions
            }

    @property
    def action_durations(self) -> dict:
        """execution durations for the prescribed actions"""
        if self.triggered is True:
            return {
                action.name: action.action_duration for action in self.actions
            }

    @property
    def started_timestamps(self) -> Dict[str, datetime.datetime]:
        """timestamps for when actions were started"""
        if self.triggered is True:
            return {
                action.name: action.started_timestamp for action in self.actions
            }

    @property
    def relative_start_times(self) -> dict:
        """relative start times for actions"""
        if self.triggered is False:
            raise ValueError(f'the {self.__class__.__name__} has not been triggered yet')
        if self.sequence_time_started is None:
            raise ValueError(f'sequence_start_time is not set for the {self.__class__.__name__} instance')
        return {
            action.name: action.time_started - self.sequence_time_started for action in self.actions
        }

    @property
    def action_returns(self) -> dict:
        """retrieves the returned values from the actions. Only actions which have completed will have return values"""
        return {
            action.name: action.method_return for action in self.actions
        }

    @property
    def actions_status(self) -> dict:
        """status of the actions of the instance"""
        return {action.name: action.status for action in self.actions}

    def wait_for_completion(self, cycle_time: float = 0.1):
        """
        Waits for the completion of the time point

        :param cycle_time: cycle check time
        """
        while self.in_progress is True:
            time.sleep(cycle_time)

    def as_dict(self) -> dict:
        """Returns a key: value representation of the instance"""
        out = self.basic_dict
        out['triggered'] = self.triggered
        out['actions'] = [
            action.as_dict() for action in self.actions
        ]
        # todo decide how to include relative start times
        return out




import time

one_second_return = 'abcde'
fifty_return = [1, 2, 3, 4, 5]


def one_second():
    time.sleep(1)
    return one_second_return


def fifty_ms():
    time.sleep(0.05)
    return fifty_return


def user_time(dur, **kwargs):
    time.sleep(dur)
    return kwargs


def five_second():
    time.sleep(5.)
    return fifty_return


def erronious():
    raise ValueError('you did something horrific and it broke the code')


def passthrough(*args, **kwargs):
    return args, kwargs



import logging
from typing import List, Union, Callable
from .action import Action, TrackedAction, ConfiguredAction
from .step import Step

logger = logging.getLogger(__name__)


def build_automation_sequence(*steps: Union[str, Step, Callable, Action, ConfiguredAction]) -> List[Step]:
    """
    Builds an automation sequence of steps based on the order provided.

    :param steps: steps to link
    :return: built automation sequence of steps
    """
    out: List[Step] = []
    for step in steps:
        # if not a step, create a step instance
        if isinstance(step, Step) is False:
            if type(step) is str:  # catch string references to actions (will only work if actions are pre-registered)
                step = Action.register_action(step)
            step = Step(step)
        # if there is a previous step, set next step attribute
        if len(out) > 0:
            out[-1].next_step = step
        out.append(step)
    return out


class Automation(ConfiguredAction):
    """
    An Automation is used to run a sequence of Steps. For an Automation, you need to specify what the first step to
    run is.

    An Automation instance is callable. When called, it will run the first Step, get the next_step from
    that first step, and loop through running and retrieving the next Step to run, until there are no more Steps to run.

    The IfStep enables decision making in an Automation

    The Automation class is also important, because all the arguments for all the Steps must be passed into the
    Automation on instantiation, and these arguments may potentially be updated by the Steps. All these arguments for
    the CustomSteps must be passed into an Automation sequence as a dictionary, where the keys are the argument names
    and the values are the initial values for these arguments. These arguments get stored as self.kwargs in the
    Automation instance, and self.kwargs gets used as the arguments to run the CustomSteps. If the CustomStep run
    function has a return value, it MUST be a dictionary, where the keys are argument names (that must correspond to
    a key in the self.kwargs dictionary of the Automation instance), and the values are the new values for the
    arguments; and in this way the arguments for the Automation instance can be updated.

    In short for using the Automation and Steps classes:
    * An Automation instantiation MUST include a dictionary of arguments and initial values to be used by the Steps
        in the Automation, unless no variables need to be passed between/used by multiple Steps
    * The parameter names for all the functions of CustomSteps must be identical to the keys in the kwargs
        dictionary that get passed into an Automation instance on instantiation
    * The return value of any CustomStep must be a dictionary, with the keys being identical to the keys in the
        kwargs dictionary of the Automation instance, and the values are the updated values
    * All these functions for the CustomSteps must also have a **kwargs parameter, to catch arguments that get passed
        through that are not actually required by the functions.
    * An Automation instance must know what the first step to run is
    * Steps need to be 'chained' to each other

    """
    def __init__(self,
                 first_step: Step,
                 **kwargs):
        """

        :param Step, first_step: first step in the automation to run
        :param dict, kwargs: all the entry keys need to be identical to the parameter names for the callable actions
            for all the Steps in the automation
        """
        if isinstance(first_step, Step) is False:
            raise TypeError('first_step must be of type Step')
        self._executions: List[TrackedAction] = []
        self._first_step: Step = first_step
        self._next_step = None
        self.kwargs = kwargs

        super().__init__(action=self.run_automation,
                         **kwargs,
                         )

        self.logger = logger.getChild(f'{self.__class__.__name__} {self.name}')

    def __call__(self, **kwargs):
        """
        Run the first step in the automation, and get the next step that results from it. then loop, running the next
        step after that and so on, until the end of the automation has been reached.

        :param dict, kwargs: dictionary of arguments to run the Steps in the automation. Nothing should need to be
            passed in here for this because the kwargs should have been passed in when the Automation instance was
            instantiated, but if they were not or need to be updated, they can be passed here
        :return:
        """
        self._executions.append(
            TrackedAction.create_from_configured(self.action)
        )
        self._executions[-1].__call__(**kwargs)

    @property
    def executions(self) -> List[TrackedAction]:
        """list of automation executions"""
        return self._executions

    @property
    def first_step(self) -> Step:
        """The first step in the Automation run"""
        return self._first_step

    @first_step.setter
    def first_step(self,
                   value: Step,
                   ):
        if isinstance(value, Step) is False:
            raise TypeError('value must be of type Step')

        self._first_step = value

    @property
    def next_step(self) -> Step:
        """The next step in the Automation run"""
        return self._next_step

    @next_step.setter
    def next_step(self,
                  value: Step
                  ):
        if isinstance(value, Step) is False:
            raise TypeError('value must be of type Step')

        self._next_step = value

    def run_step_and_return_next_step(self,
                                      step: Step,
                                      **kwargs,
                                      ) -> Step:
        """
        Run a Step, and get the next_step from the Step that was run and return it. If the Step is a CustomStep,
        a dictionary will be returned when the step is run. The dictionary will have identical keys to the kwargs for
        the Automation instance. If a dictionary is returned, then update self.kwargs for the Automation instance
        with the return from the Step

        :param Step, step: a Step to run
        :param dict, kwargs: if the Step callable (action) requires arguments, they must be passed in through this
            kwargs dictionary
        :return: Step, self._next_step: the next Step that should be run
        """
        kwargs_from_step_that_was_run = step(**kwargs)  # note that the next step may be set within the run function
        #  in some of the subclasses or depending on what the run function is

        if type(kwargs_from_step_that_was_run) == dict:
            # if the return type of the run function is a dictionary, then update self.kwargs of the automation object
            self.kwargs.update(kwargs_from_step_that_was_run)

        self._next_step = step.next_step
        return self._next_step

    def run_automation(self,
                       **kwargs,
                       ):
        """
        Run the first step in the automation, and get the next step that results from it. then loop, running the next
        step after that and so on, until the end of the automation has been reached. Will get called when an
        Automation instance is called.

        :param dict, kwargs: dictionary of arguments to run the Steps in the automation. Nothing should need to be
            passed in here for this because the kwargs should have been passed in when the Automation instance was
            instantiated, but if they were not or need to be updated, they can be passed here
        :return:
        """
        self.logger.debug(f'run automation')
        self.kwargs.update(kwargs)
        self._next_step = self._first_step
        while self._next_step is not None:
            self.logger.debug(f'next step: {self._next_step}')
            self._next_step = self.run_step_and_return_next_step(step=self._next_step, **self.kwargs)
        self.logger.debug('automation ended')




from .version import __version__
from .action import Action, TrackedAction
from .step import Step, IfStep, CustomStep, ConditionCheck
from .timepoint import TimePoint, ActionTimePoint
from .automation import Automation
from .scheduler import SamplingScheduler


__all__ = [
    'SamplingScheduler',
    'Action',
    'TrackedAction',
    'Automation',
    'TimePoint',
    'ActionTimePoint',
    'Step',
    'IfStep',
    'CustomStep',
    'ConditionCheck'
]



import unittest
from hein_control.action import Action, TrackedAction
from .imported_methods import *


class TestSchedulerAction(unittest.TestCase):
    def setUp(self) -> None:
        # register all actions
        Action.register_action(one_second)
        Action.register_action(fifty_ms)
        Action.register_action(user_time)

    def test_registry(self):
        """tests registry functionality of SchedulerAction"""
        self.assertTrue(
            Action.action_is_registered(one_second)
        )
        self.assertTrue(
            Action.action_is_registered(fifty_ms)
        )

    def test_scheduler_action(self):
        """test functionality of action calling and return"""
        action = Action.register_action(one_second)
        self.assertEqual(  # test name attribute
            action.name,
            one_second.__name__
        )
        ret = action()  # execute action, test return
        self.assertEqual(
            one_second_return,
            ret
        )

    def test_argument_passthrough(self):
        """tests argument and kwarg passthrough"""
        action = Action.register_action(passthrough)
        args = (0.123, 123, 456)
        kwargs = {'asdf': 'qwerty'}
        ret = action(
            *args,
            **kwargs
        )
        self.assertEqual(
            (args, kwargs),
            ret
        )

    def test_arg_kwarg(self):
        """tests argument and kwarg specification and passthrough"""
        args = (0.123, 123, 456)
        kwargs = {'asdf': 'qwerty'}
        # test argument updating
        action = TrackedAction(
            passthrough,
            *args,
            **kwargs
        )
        self.assertEqual(
            action.args,
            args
        )
        self.assertEqual(
            action.kwargs,
            kwargs
        )
        action(
            789,
            asdf='al;skdjf;'
        )
        self.assertEqual(
            action.args,
            (789, 123, 456)
        )
        self.assertEqual(
            action.kwargs,
            {'asdf': 'al;skdjf;'}
        )

    def test_error_catching(self):
        action = Action.register_action(erronious)
        self.assertRaises(
            ValueError,
            action
        )
        self.assertTrue(action.error)
        self.assertIsNotNone(action.error_details)


class TestTrackedAction(unittest.TestCase):
    def test(self):
        tracked_action = TrackedAction(
            one_second
        )
        start_time = time.time()
        tracked_action()
        end_time = time.time()
        self.assertAlmostEqual(
            start_time,
            tracked_action.time_started
        )
        self.assertAlmostEqual(
            end_time,
            tracked_action.time_completed
        )
        self.assertAlmostEqual(
            end_time - start_time,
            tracked_action.action_duration
        )
        self.assertEqual(
            tracked_action.name,
            one_second.__name__
        )
        self.assertEqual(
            tracked_action.method_return,
            one_second_return
        )
        dct = tracked_action.as_dict()  # todo more elaborate test for dict testing



import time
import datetime
import copy
from typing import List, Union, Callable


# todo figure out how to trigger a pre-trigger event (e.g. turn on instrument and condition)
#   - executed contextually
#   - or rely on the user to specify pre-events in provided methods

# dictionary of action states
ACTION_STATES = {
    -1: 'ERROR',
    0: 'CONFIGURED',
    1: 'PENDING',
    2: 'EXECUTING',
    3: 'COMPLETE',
}


class Action:
    # registry of defined actions
    _registry: List["Action"] = []

    def __init__(self,
                 action: Callable,
                 *default_args,
                 **default_kwargs,
                 ):
        """
        Class for storing and tracking registered methods.

        :param action: action method to execute when called
        :param default_args: default arguments for the action
        :param default_kwargs: default kwargs for the action
        """
        # todo consider implementing action removal
        if self.action_is_registered(action) is True:
            raise ValueError(
                f'The action "{action}" is already registered, use SchedulerAction.register_action instead'
            )
        self.action = action
        # error state tracker
        self.error: bool = None
        self.error_details = None
        # todo create properties and method for adding/removing arguments and kwargs
        self.default_args = default_args
        self.default_kwargs = default_kwargs
        Action._registry.append(self)
        # todo retrieve and store docstring of function

    def __repr__(self):
        return f'{self.__class__.__name__}({self.action.__name__})'

    def __call__(self, *args, **kwargs):
        try:
            ret = self.action(*args, **kwargs)
            self.error = False  # signify successful execution
            return ret
        except Exception as e:
            self.error = True
            self.error_details = e
            raise e

    @property
    def name(self) -> str:
        """action method name"""
        return self.action.__name__

    @classmethod
    def action_is_registered(cls,
                             action: Union[str, Callable, "Action"]
                             ) -> bool:
        """
        Checks whether the provided action is registered as a SchedulerAction instance.

        :param action: action to check
        :return: whether action is registered
        """
        for instance in cls._registry:
            if any([
                instance.name == action,
                instance == action,
                instance.action == action,
            ]):
                return True
        return False

    @classmethod
    def register_action(cls,
                        action: Union[str, Callable, "Action"],
                        ) -> "Action":
        """
        Registers a new instance or returns an existing instance if the action is already registered.

        :param action: action method
        :return: Action instance
        """
        # todo catch, passthrough, or assign args and kwargs
        # todo consider logging that a new action has been registered
        # catch class instance
        if isinstance(action, Action):
            return action
        elif hasattr(action, 'action') and isinstance(action.action, Action):
            return action.action
        # search for action name
        elif type(action) is str:
            for instance in cls._registry:
                if instance.name == action:
                    return instance
        else:
            # check for preexisting actions
            for instance in cls._registry:
                if instance.name == action.__name__:
                    return instance
            # if not found, create and return
            return cls(
                action
            )

    @classmethod
    def register_actions(cls, *actions: Union[str, Callable, "Action"]) -> List["Action"]:
        """
        Registers multiple actions

        :param actions: action methods to be registered
        :return: list of actions which were registered
        """
        return [
            cls.register_action(action) for action in actions
        ]

    @classmethod
    def get_registered_actions(cls) -> List["Action"]:
        """returns a list of registered Scheduler Actions"""
        return cls._registry

    @classmethod
    def get_registered_action_names(cls) -> List[str]:
        """returns a list of the names of the registered Scheduler Actions"""
        return [action.name for action in cls._registry]


def _update_args(original_args: tuple, new_args: tuple) -> tuple:
    """
    Updates the original argument tuple with the new arguments.

    :param original_args: original argument tuple
    :param new_args: new argument tuple
    :return: updated argument tuple
    """
    original_args = list(original_args)
    for ind, val in enumerate(new_args):
        try:
            original_args[ind] = val
        except IndexError:
            original_args.append(val)
    return tuple(original_args)


def is_builtin_type(obj):
    """returns whether the object is a builtin type"""
    return obj.__class__.__module__ == 'builtins'


def _update_kwargs(original_kwargs: dict, new_kwargs: dict) -> dict:
    """
    Updates the original kwargs with the new kwargs. Prevents mutation of the original kwargs.

    Only perform deepcopy on builtin types and interface instances that have specified the deepcopy method.

    :param original_kwargs: original keyword arguments
    :param new_kwargs: new keyword arguments
    :return: consolidated and updated keyword arguments
    """
    # create deepcopies of supporting types
    dct = {
        key: copy.deepcopy(value)
        for key, value in original_kwargs.items()
        if hasattr(value, '__deepcopy__') or is_builtin_type(value)
    }
    # add direct references for non-copyable types
    dct.update({
        key: original_kwargs[key] for key in original_kwargs.keys() - dct.keys()
    })
    # finally update with new kwargs
    dct.update(new_kwargs)
    return dct


class ConfiguredAction:
    def __init__(self,
                 action: Union[Callable, Action, str],
                 *args,
                 **kwargs,
                 ):
        """
        An action that is configured for execution

        :param action: action to perform (callable)
        :param args: arguments for the action
        :param kwargs: keyword arguments for the action
        """
        self._action: Action = None
        # register action and store attribute
        self.action = action
        # todo inspect provided action for default kwargs and store values
        self.args = _update_args(self.action.default_args, args)
        self.kwargs = _update_kwargs(self.action.default_kwargs, kwargs)

    def __str__(self):
        return f'{self.__class__.__name__} {self.name}'

    def __repr__(self):
        return self.__str__()

    def __call__(self, *args, **kwargs):
        raise NotImplementedError(f'{self.__class__.__name__} instances may not be called')

    def __copy__(self):
        return self.__init__(
            self.action,
            self.args,
            self.kwargs,
        )

    def __deepcopy__(self, memodict={}):
        return self.__copy__()

    @property
    def action(self) -> Action:
        """The action to be tracked"""
        return self._action

    @action.setter
    def action(self, value: Union[Callable, Action, str]):
        if value is not None:
            self._action = Action.register_action(value)

    @property
    def name(self) -> str:
        """method name"""
        return self.action.name

    def update_args(self, *args):
        """Updates the arguments for the Action"""
        self.args = _update_args(
            self.args,
            args
        )

    def update_kwargs(self, **kwargs):
        """updates the keyword arguments for the Action"""
        self.kwargs = _update_kwargs(
            self.kwargs,
            kwargs
        )

    @property
    def status_code(self) -> int:
        """returns a status code representative of the state of the Action"""
        return 0

    @property
    def status_string(self) -> str:
        """string meaning of status code"""
        return ACTION_STATES[self.status_code]

    @property
    def status(self) -> str:
        """status string for the Action"""
        return f'{self.status_code}: {self.status_string}'

    def as_dict(self) -> dict:
        """dictionary of relevant information"""
        out = {
            'name': self.name,
            'arguments': self.args,
            'keyword_arguments': self.kwargs,
            'status_code': self.status_code,
            'status': self.status_string,
        }
        return out


class TrackedAction(ConfiguredAction):
    def __init__(self,
                 action: Union[Callable, ConfiguredAction],
                 *args,
                 **kwargs,
                 ):
        """
        Creates a wrapped action which will track the time of start, completion, and return of the action.

        :param action: action to perform (callable)
        :param args: arguments for the action
        :param kwargs: keyword arguments for the action
        """
        # if provided with a configured action, extract parameters
        if isinstance(action, ConfiguredAction):
            args = action.args
            kwargs = action.kwargs
            action = action.action
        ConfiguredAction.__init__(
            self,
            action,
            *args,
            **kwargs,
        )
        self._time_started = None
        self._time_completed = None
        self._action_return = None

    def __call__(self, *args, **kwargs):
        if self._time_completed is not None:
            raise RuntimeError(f'multiple executions of a {self.__class__.__name__} instance is not permitted')
        # update args and kwargs and store
        self.update_args(*args)
        self.update_kwargs(**kwargs)
        self._time_started = time.time()  # set start time
        self._action_return = self.action(
            *self.args,
            **self.kwargs
        )
        self._time_completed = time.time()  # save completed time

    @property
    def time_started(self) -> float:
        """time stamp when the action was started"""
        return self._time_started

    @property
    def time_completed(self) -> float:
        """time stamp when the action was completed"""
        return self._time_completed

    @property
    def started_timestamp(self) -> datetime.datetime:
        """timestamp for when the action was started"""
        if self.time_started is not None:
            return datetime.datetime.fromtimestamp(self.time_started)

    @property
    def action_duration(self) -> float:
        """task duration (s)"""
        if self._time_completed is not None:
            return self._time_completed - self._time_started

    @property
    def method_return(self):
        """the return of the method once complete"""
        return self._action_return

    @property
    def status_code(self) -> int:
        """returns a status code representative of the state of the Action"""
        if self.action.error is True:
            code = -1
        elif self.time_started is None:
            code = 1
        elif self.time_completed is None:
            code = 2
        else:
            code = 3
        return code

    def as_dict(self) -> dict:
        """dictionary of relevant information"""
        out = {
            'name': self.name,
            'arguments': self.args,
            'keyword_arguments': self.kwargs,
            'status_code': self.status_code,
            'status': self.status_string,
        }
        if self.time_started is not None:
            out['time_started'] = self.time_started
            out['timestamp'] = str(self.started_timestamp)
        if self.time_completed is not None:
            out['time_completed'] = self.time_completed
            out['duration'] = self.time_completed - self.time_started
            if self.action.error is True:
                out['error_during_execution'] = True
                out['error_details'] = self.action.error_details
        return out

    @classmethod
    def create_from_configured(cls, source: Union[Action, ConfiguredAction]) -> "TrackedAction":
        """
        Creates a TrackedAction instance from a provided Action or ConfiguredAction

        :param source: action to generate from
        :return: instantiated TrackedAction
        """
        if isinstance(source, Action):
            args = ()
            kwargs = {}
            action = source.action
        elif isinstance(source, ConfiguredAction):
            args = source.args
            kwargs = source.kwargs
            action = source.action
        else:
            raise TypeError(f'the provided source is not an interpretable type: {type(source)}')
        out = cls(action=action)
        out.update_args(*args)
        out.update_kwargs(**kwargs)
        return out



import logging
from hein_control.automation import Automation
from hein_control.step import CustomStep, ConditionCheck, IfStep

class AA():
    def __init__(self):
        self.logger = logger.getChild(f'{self.__class__.__name__}')

    def func_a(self):
        self.logger.debug('func_a')

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)

aa = AA()

automation_kwargs = {
    'aa': aa,
    'a': 'a',
}

def action_1(aa, a, **kwargs):
    aa.func_a()
    return

def action_2(**kwargs):
    return

def condition_1(**kwargs):
    return True


step_a = CustomStep(action=action_1)
step_b = CustomStep(action=action_2)

cond_a = ConditionCheck(condition=condition_1, next_step=step_a)

if_check = IfStep(condition_check_list=[cond_a,
                                        ])

step_a.next_step = step_b
step_b.next_step = if_check

automation = Automation(first_step=step_a,
                        **automation_kwargs)

automation()